using System;
using System.Collections.Generic;
using System.Linq;
using Battlehub.RTSaveLoad;
using MacGruber;
using SimpleJSON;
using UnityEngine;
using UnityEngine.Serialization;

namespace CheesyFX
{
    public class Fuckable : MonoBehaviour
    {
        public new string name;
        protected static List<object> UIElements = new List<object>();
        // public JSONStorableBool autoThrust = new JSONStorableBool("Thrust", true);
        // public JSONStorableBool maleThrust = new JSONStorableBool("Male Thrust", false);
        public JSONStorableString info = new JSONStorableString("info", "Idle");
        public JSONStorableFloat depth = new JSONStorableFloat("Depth", 0f, 0f, .4f, true, false);
        public JSONStorableFloat speed = new JSONStorableFloat("Speed", 0f, -5f, 5f, true, false);
        
        public JSONStorableFloat sensitivity = new JSONStorableFloat("Sensitivity (ReadMyLips)", 1f, 0f, 5f);
        public JSONStorableFloat penetrationSoundsVolume = new JSONStorableFloat("Penetration Sounds Volume", 1f, 0f, 1f);
        
        public JSONStorableBool magnetic = new JSONStorableBool("Magnetic", true);
        
        protected FloatTriggerManager depthTriggers;
        protected FloatTriggerManager speedTriggers;
        
        protected UIDynamicButton depthTriggerButton;
        protected UIDynamicButton speedTriggerButton;
        
        public Transform enterPointTF;
        public CapsuleCollider proximityTrigger;
        public AudioSource audioSource;
        
        public Penetrator penetrator;
        public Transform forceTarget;
        public Force thrustForce;
        // public Force depthForce;
        
        public Force maleForce;
        public Vector3 forward;
        public Vector3 rootToOrifice;
        protected Rigidbody rb;

        public Magnet magnet;

        public virtual void Init(string name)
        {
            this.name = name;
            depthTriggers = FillMeUp.singleton.AddFloatTriggerManager(depth, false, 0f, 0f, .25f);
            speedTriggers = FillMeUp.singleton.AddFloatTriggerManager(speed, true, 0f, 0f, .5f);
            if (this is Hand) rb = FillMeUp.atom.rigidbodies.First(x => x.name == name.First() + "Hand");
            else if(this is Throat) rb = FillMeUp.atom.rigidbodies.First(x => x.name == "head");
            else rb = FillMeUp.atom.rigidbodies.First(x => x.name == "hip");
            thrustForce = gameObject.AddComponent<Force>().Init(name+":Thrust", rb, () => forward);
            thrustForce.enabledJ.setCallbackFunction += val =>
            {
                thrustForce.SetActive(val && enabled);
            };
            maleForce = gameObject.AddComponent<Force>().Init($"{name}:MaleThrust", null, () => rootToOrifice);
            maleForce.control.offset.val = 200f;
            maleForce.amplitude.mean.SetWithDefault(300f);
            maleForce.amplitude.delta.SetWithDefault(200f);
            maleForce.period.mean.SetWithDefault(.5f);
            maleForce.enabledJ.setCallbackFunction += val => maleForce.SetActive(val && enabled);
            
            FillMeUp.forceChooser.choices.Add(thrustForce.name);
            FillMeUp.forceChooser.choices.Add(maleForce.name);
        }

        public virtual void FixedUpdate()
        {
            forward = -penetrator.forward();
            rootToOrifice = (enterPointTF.position - forceTarget.position).normalized;
        }

        public void ClearUI()
        {
            FillMeUp.singleton.RemoveUIElements(UIElements);
        }
        
        public virtual void SelectTab(int id)
        {
            FillMeUp.singleton.RemoveUIElements(UIElements);
            if(id < 4) FillMeUp.debugMode = false;
            if (id == 0) CreateTriggersUI();
            else if(id == 1) CreateSettingsUI();
            else if(id == 2) CreateBulgeUI();
            else if(id == 3) CreateForcesUI();
            else if(id == 4) CreateDebugUI();
        }
        
        public virtual JSONClass Store(string subScenePrefix, bool storeTriggers = true)
        {
            JSONClass jc = new JSONClass();
            magnetic.Store(jc);
            sensitivity.Store(jc);
            penetrationSoundsVolume.Store(jc);
            // thrustForce.enabledJ.Store(jc);
            jc["Thrust"] = thrustForce.Store;
            if (maleForce) jc["MaleThrust"] = maleForce.Store;
            // jc["DepthForce"] = depthForce.Store;
            if(storeTriggers)
            {
                jc["DepthTriggers"] = depthTriggers.Store();
                jc["SpeedTriggers"] = speedTriggers.Store();
            }
            return jc;
        }
        
        public virtual void Load(JSONClass jc, string subScenePrefix)
        {
            if(jc.HasKey(name))
            {
                JSONClass tc = jc[name].AsObject;
                magnetic.Load(tc);
                sensitivity.Load(tc);
                penetrationSoundsVolume.Load(tc);
                // autoThrust.Load(tc);
                thrustForce.Load(tc["Thrust"].AsObject);
                if(maleForce) maleForce.Load(tc["MaleThrust"].AsObject);
                // depthForce.Load(tc["DepthForce"].AsObject);
                depthTriggers.Load(tc["DepthTriggers"].AsObject);
                speedTriggers.Load(tc["SpeedTriggers"].AsObject);
            }
            SyncTriggerButtons();
        }
        
        public virtual void SyncTriggerButtons()
        {
            depthTriggerButton.label = $"Depth Triggers ({depthTriggers.triggers.Count})";
            speedTriggerButton.label = $"Speed Triggers ({speedTriggers.triggers.Count})";
        }

        public virtual void CreateTriggersUI()
        {
            depthTriggerButton = FillMeUp.singleton.CreateButton($"Depth Triggers ({depthTriggers.triggers.Count})");
            depthTriggerButton.buttonColor = new Color(0.45f, 1f, 0.45f);
            depthTriggerButton.button.onClick.AddListener(() =>
            {
                {
                    FillMeUp.singleton.ClearUI();
                    depthTriggers.OpenPanel(FillMeUp.singleton, FillMeUp.singleton.CreateUI);
                }
            });
            UIElements.Add(depthTriggerButton);
            depth.CreateUI(FillMeUp.singleton, UIElements:UIElements);
			     
            speedTriggerButton = FillMeUp.singleton.CreateButton($"Speed Triggers ({speedTriggers.triggers.Count})", true);
            speedTriggerButton.buttonColor = new Color(0.45f, 1f, 0.45f);
            speedTriggerButton.button.onClick.AddListener(() =>
            {
                {
                    FillMeUp.singleton.ClearUI();
                    speedTriggers.OpenPanel(FillMeUp.singleton, FillMeUp.singleton.CreateUI);
                }
            });
            UIElements.Add(speedTriggerButton);
            speed.CreateUI(FillMeUp.singleton, UIElements:UIElements, rightSide:true);
            var textField = FillMeUp.singleton.CreateTextField(info, true);
            UIElements.Add(textField);
            textField.height = 50f;
        }

        public virtual void CreateSettingsUI()
        {
            
        }

        public virtual void CreateBulgeUI()
        {
            
        }

        protected static string toggleInfoText =
            "You can toggle all forces between 'off' and the state they were before by" +
            " pressing '<b>CTRL+F</b>' or calling the action '<b>Toggle Forces</b>' via trigger.\n" +
            "If you never want forces setup a FillMeUp preset called '<b>UserDefaults</b>'.";

        public virtual void CreateForcesUI()
        {
            thrustForce.enabledJ.CreateUI(UIElements);
            maleForce.enabledJ.CreateUI(UIElements);
            magnetic.CreateUI(UIElements);
            FillMeUp.singleton.SetupButton("Configure Thrust", true, () => CreateForceUINewPage(thrustForce), UIElements);
            FillMeUp.singleton.SetupButton("Configure Male Thrust", true, () => CreateForceUINewPage(maleForce), UIElements);
            if (magnet != null)
            {
                FillMeUp.singleton.SetupButton("Configure Magnet", true, magnet.CreateUINewPage, UIElements);
            }
            else
            {
                FillMeUp.singleton.SetupButton("N.A.", true, delegate {  }, UIElements);
            }
            if(this is Hand)
            {
                FillMeUp.singleton.SetupButton("Configure Hands Sync Group", true, FillMeUp.handForceGroup.CreateUI,
                    UIElements);
                FillMeUp.handForceGroup.driverInfo.CreateUI(UIElements);
            }
            else
            {
                FillMeUp.singleton.SetupButton("Configure Orifice Sync Group", true,
                    FillMeUp.orificeForceGroup.CreateUI, UIElements);
                FillMeUp.orificeForceGroup.driverInfo.CreateUI(UIElements);
            }

            var info = Utils.SetupInfoTextNoScroll(FillMeUp.singleton, toggleInfoText, 200f, false);
            info.background.offsetMin = new Vector2(0, 0);
            UIElements.Add(info);
        }

        public virtual void CreateForceUINewPage(Force force)
        {
            FillMeUp.singleton.ClearUI();
            UIElements.Clear();
            var button = FillMeUp.singleton.CreateButton("Return");
            button.buttonColor = new Color(0.55f, 0.90f, 1f);
            button.button.onClick.AddListener(
                () =>
                {
                    FillMeUp.singleton.ClearUI();
                    FillMeUp.singleton.CreateUI();
                    FillMeUp.singleton.settingsTabbar.SelectTab(3);
                    force.control.UIOpen = false;
                });
            UIElements.Add(button);
            var chooser = FillMeUp.forceChooser.CreateUI(UIElements, true, chooserType:0);
            chooser.ForceHeight(50f);
            FillMeUp.forceChooser.valNoCallback = force.name;
            force.control.CreateUI(FillMeUp.singleton);
            force.control.UIOpen = true;
        }

        public virtual void CreateDebugUI()
        {
            FillMeUp.debugMode = true;
        }

        // public virtual void OnEnable()
        // {
        //     if(autoThrust.val) thrustForce.enabled = true;
        // }
    }
}